# pylint: disable=duplicate-code
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, cast

from huggingface_hub.inference._text_generation import (  # type: ignore
    TextGenerationResponse,
    TextGenerationStreamResponse,
)
from huggingface_hub.inference._types import ConversationalOutput  # type: ignore

from nebuly.entities import HistoryEntry, ModelInput
from nebuly.providers.base import ProviderDataExtractor
from nebuly.providers.utils import get_argument


@dataclass(frozen=True)
class HuggingFaceHubDataExtractor(ProviderDataExtractor):
    original_args: tuple[Any, ...]
    original_kwargs: dict[str, Any]
    function_name: str

    def _extract_history(self) -> list[HistoryEntry]:
        generated_responses = get_argument(
            self.original_args, self.original_kwargs, "generated_responses", 2
        )
        past_user_inputs = get_argument(
            self.original_args, self.original_kwargs, "past_user_inputs", 3
        )
        history = []

        for user_input, assistant_response in zip(
            past_user_inputs if past_user_inputs is not None else [],
            generated_responses if generated_responses is not None else [],
        ):
            history.append(HistoryEntry(user=user_input, assistant=assistant_response))
        return history

    def extract_input_and_history(self) -> ModelInput:
        if self.function_name in [
            "InferenceClient.conversational",
            "AsyncInferenceClient.conversational",
        ]:
            prompt = get_argument(self.original_args, self.original_kwargs, "text", 1)
            history = self._extract_history()
            return ModelInput(prompt=prompt, history=history)
        if self.function_name in [
            "InferenceClient.text_generation",
            "AsyncInferenceClient.text_generation",
        ]:
            prompt = get_argument(self.original_args, self.original_kwargs, "prompt", 1)
            return ModelInput(prompt=prompt)

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self,
        stream: bool,
        outputs: str
        | ConversationalOutput
        | TextGenerationResponse
        | list[str]
        | list[TextGenerationStreamResponse],
    ) -> str:
        if stream and isinstance(outputs, list):
            return self._extract_output_generator(outputs)

        if self.function_name in [  # ConversationalOutput
            "InferenceClient.conversational",
            "AsyncInferenceClient.conversational",
        ]:
            result = cast(ConversationalOutput, outputs)
            return cast(str, result["generated_text"])
        if isinstance(
            outputs, (str, TextGenerationResponse)
        ) and self.function_name in [
            "InferenceClient.text_generation",
            "AsyncInferenceClient.text_generation",
        ]:
            if isinstance(outputs, TextGenerationResponse):
                return cast(str, outputs.generated_text)
            return outputs

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"output type: {type(outputs)}"
        )

    def _extract_output_generator(
        self, outputs: list[str] | list[TextGenerationStreamResponse]
    ) -> str:
        if self.function_name in [
            "InferenceClient.text_generation",
            "AsyncInferenceClient.text_generation",
        ]:
            if all(isinstance(output, str) for output in outputs):
                result = cast(List[str], outputs)
                return "".join(result)
            if all(
                isinstance(output, TextGenerationStreamResponse) for output in outputs
            ):
                result = cast(List[TextGenerationStreamResponse], outputs)
                return "".join(
                    [
                        output.generated_text  # type: ignore
                        if output.generated_text is not None  # type: ignore
                        else ""
                        for output in result
                    ]
                )

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"output type: {type(outputs)}"
        )



from __future__ import annotations

import os
from functools import partial
from queue import Queue

from nebuly.config import PACKAGES
from nebuly.consumers import ConsumerWorker
from nebuly.entities import InteractionWatch, Observer
from nebuly.exceptions import APIKeyNotProvidedError, NebulyAlreadyInitializedError
from nebuly.monkey_patching import (
    check_no_packages_already_imported,
    import_and_patch_packages,
)
from nebuly.observers import NebulyObserver
from nebuly.requests import post_message

_initialized = False


def init(
    *,
    api_key: str | None = None,
    anonymize: bool = True,
) -> None:
    if not api_key:
        api_key = _get_api_key()

    _check_nebuly_is_not_initialized()
    check_no_packages_already_imported(PACKAGES)
    observer = _create_observer_and_start_publisher(
        api_key=api_key, anonymize=anonymize
    )
    import_and_patch_packages(PACKAGES, observer)

    try:
        from nebuly.tracking_handlers import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            set_tracking_handlers,
        )

        set_tracking_handlers()
    except ImportError:
        pass


def _get_api_key() -> str:
    api_key = os.environ.get("NEBULY_API_KEY")
    if not api_key:
        raise APIKeyNotProvidedError(
            "API key not provided and not found in environment"
        )
    return api_key


def _check_nebuly_is_not_initialized() -> None:
    global _initialized  # pylint: disable=global-statement
    if _initialized:
        raise NebulyAlreadyInitializedError("Nebuly already initialized")
    _initialized = True


def _create_observer_and_start_publisher(
    *, api_key: str, anonymize: bool = True
) -> Observer:
    queue: Queue[InteractionWatch] = Queue()

    ConsumerWorker(queue, partial(post_message, api_key=api_key, anonymize=anonymize))
    observer = NebulyObserver(
        api_key=api_key,
        publish=queue.put,
    )
    return observer.on_event_received



from __future__ import annotations

import logging
from typing import Any, Sequence
from uuid import UUID

from langchain.callbacks.base import BaseCallbackHandler, BaseCallbackManager, Callbacks
from langchain.chains.base import Chain
from langchain.schema import Document
from langchain.schema.messages import BaseMessage
from langchain.schema.output import LLMResult

from nebuly.contextmanager import EventData, EventsStorage, get_nearest_open_interaction
from nebuly.entities import EventType

logger = logging.getLogger(__name__)


def set_tracking_handlers() -> None:
    tracking_handler = LangChainTrackingHandler()
    original_call = Chain.__call__
    original_acall = Chain.acall

    def set_callbacks_arg(callbacks: Callbacks) -> Callbacks:
        if callbacks is None:
            callbacks = [tracking_handler]
        elif isinstance(callbacks, list):
            if tracking_handler not in callbacks:
                callbacks.append(tracking_handler)
        elif isinstance(callbacks, BaseCallbackManager):
            if tracking_handler not in callbacks.handlers:
                callbacks.handlers.append(tracking_handler)
            if tracking_handler not in callbacks.inheritable_handlers:
                callbacks.inheritable_handlers.append(tracking_handler)
        return callbacks

    def tracked_call(
        self: Any,
        inputs: dict[str, Any] | Any,
        return_only_outputs: bool = False,
        callbacks: Callbacks = None,
        **kwargs: Any,
    ) -> Any:
        callbacks = set_callbacks_arg(callbacks)
        if isinstance(inputs, dict):
            tracking_handler.nebuly_user = inputs.pop("user_id", None)
            tracking_handler.nebuly_user_group = inputs.pop("user_group_profile", None)
        if tracking_handler.nebuly_user is None:
            tracking_handler.nebuly_user = kwargs.pop("user_id", None)
        if tracking_handler.nebuly_user_group is None:
            tracking_handler.nebuly_user_group = kwargs.pop("user_group_profile", None)
        return original_call(
            self,
            inputs=inputs,
            return_only_outputs=return_only_outputs,
            callbacks=callbacks,
            **kwargs,
        )

    def tracked_acall(
        self: Any,
        inputs: dict[str, Any],
        return_only_outputs: bool = False,
        callbacks: Callbacks = None,
        **kwargs: Any,
    ) -> Any:
        callbacks = set_callbacks_arg(callbacks)
        tracking_handler.nebuly_user = kwargs.pop("user_id", None)
        tracking_handler.nebuly_user_group = kwargs.pop("user_group_profile", None)
        return original_acall(
            self,
            inputs=inputs,
            return_only_outputs=return_only_outputs,
            callbacks=callbacks,
            **kwargs,
        )

    Chain.__call__ = tracked_call  # type: ignore
    Chain.acall = tracked_acall  # type: ignore


class LangChainTrackingHandler(BaseCallbackHandler):  # noqa
    def __init__(self) -> None:
        self.nebuly_user = None
        self.nebuly_user_group = None

    @property
    def current_interaction_storage(self) -> EventsStorage:
        return (
            get_nearest_open_interaction()._events_storage  # pylint: disable=protected-access  # noqa: E501
        )

    def on_tool_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        input_str: str,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.TOOL,
            kwargs={
                "serialized": serialized,
                "input_str": input_str,
                **kwargs,
            },
        )
        self.current_interaction_storage.add_event(
            run_id, parent_run_id, data, module="langchain"
        )

    def on_tool_end(  # pylint: disable=arguments-differ
        self,
        output: str,
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        self.current_interaction_storage.events[run_id].data.add_end_event_data(
            kwargs=kwargs, output=output
        )
        self.current_interaction_storage.events[run_id].set_end_time()

    def on_retriever_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        query: str,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.RETRIEVAL,
            kwargs={
                "serialized": serialized,
                "query": query,
                **kwargs,
            },
        )
        self.current_interaction_storage.add_event(
            run_id, parent_run_id, data, module="langchain"
        )

    def on_retriever_end(  # pylint: disable=arguments-differ
        self,
        documents: Sequence[Document],
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        self.current_interaction_storage.events[run_id].data.add_end_event_data(
            kwargs=kwargs, output=documents
        )
        self.current_interaction_storage.events[run_id].set_end_time()

    def on_llm_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        prompts: list[str],
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.LLM_MODEL,
            kwargs={
                "serialized": serialized,
                "prompts": prompts,
                **kwargs,
            },
        )
        self.current_interaction_storage.add_event(
            run_id, parent_run_id, data, module="langchain"
        )

    def on_llm_end(  # pylint: disable=arguments-differ
        self,
        response: LLMResult,
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        self.current_interaction_storage.events[run_id].data.add_end_event_data(
            kwargs=kwargs, output=response
        )
        self.current_interaction_storage.events[run_id].set_end_time()

    def on_chat_model_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        messages: list[list[BaseMessage]],
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.CHAT_MODEL,
            kwargs={
                "serialized": serialized,
                "messages": messages,
                **kwargs,
            },
        )
        self.current_interaction_storage.add_event(
            run_id, parent_run_id, data, module="langchain"
        )

    def on_chain_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        inputs: dict[str, Any],
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.CHAIN,
            kwargs={
                "serialized": serialized,
                "inputs": inputs,
                **kwargs,
            },
        )
        self.current_interaction_storage.add_event(
            run_id, parent_run_id, data, module="langchain"
        )

    def on_chain_end(  # pylint: disable=arguments-differ
        self,
        outputs: dict[str, Any],
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        self.current_interaction_storage.events[run_id].data.add_end_event_data(
            kwargs=kwargs, output=outputs
        )
        self.current_interaction_storage.events[run_id].set_end_time()



from __future__ import annotations

import atexit
from queue import Empty, Queue
from threading import Thread

from nebuly.entities import InteractionWatch, Publisher


class ConsumerWorker:
    def __init__(self, queue: Queue[InteractionWatch], publish: Publisher) -> None:
        self.publish = publish
        self.queue: Queue[InteractionWatch] = queue
        self.running = True
        self.thread = Thread(target=self.run, daemon=True)
        self.thread.start()
        atexit.register(self.stop)

    def run(self) -> None:
        while self.running:
            try:
                message = self.queue.get(timeout=0.1)
            except Empty:
                continue
            self.publish(message)
            self.queue.task_done()

    def stop(self) -> None:
        self.queue.join()
        self.running = False
        self.thread.join()



# pylint: disable=duplicate-code, wrong-import-position, import-error, no-name-in-module
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Callable, cast

from google.generativeai.discuss import ChatResponse  # type: ignore
from google.generativeai.text import Completion  # type: ignore
from google.generativeai.types import discuss_types, text_types  # type: ignore

from nebuly.entities import HistoryEntry, ModelInput
from nebuly.providers.base import PicklerHandler, ProviderDataExtractor
from nebuly.providers.utils import get_argument

logger = logging.getLogger(__name__)


class GooglePicklerHandler(PicklerHandler):
    @property
    def _object_key_attribute_mapping(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        completion_names = ["candidates", "result", "filters", "safety_feedback"]
        chat_names = [
            "model",
            "context",
            "examples",
            "messages",
            "temperature",
            "candidate_count",
            "candidates",
            "filters",
            "top_p",
            "top_k",
        ]

        return {
            Completion: {
                "key_names": completion_names,
                "attribute_names": completion_names,
            },
            ChatResponse: {
                "key_names": chat_names,
                "attribute_names": chat_names,
            },
        }


def handle_google_unpickable_objects() -> None:
    pickler_handler = GooglePicklerHandler()
    for obj in [Completion, ChatResponse]:
        pickler_handler.handle_unpickable_object(
            obj=obj,
        )


@dataclass(frozen=True)
class GoogleDataExtractor(ProviderDataExtractor):
    original_args: tuple[Any, ...]
    original_kwargs: dict[str, Any]
    function_name: str

    def _extract_history(self) -> list[HistoryEntry]:
        if self.function_name == "generativeai.discuss.ChatResponse.reply":
            history = getattr(self.original_args[0], "messages", [])
            history = [message["content"] for message in history]
        else:
            history = self.original_kwargs.get("messages", [])[:-1]

        if len(history) % 2 != 0:
            logger.warning("Odd number of chat history elements, ignoring last element")
            history = history[:-1]

        # Convert the history to [(user, assistant), ...] format
        history = [
            HistoryEntry(user=history[i], assistant=history[i + 1])
            for i in range(0, len(history), 2)
            if i < len(history) - 1
        ]

        return history

    def extract_input_and_history(self) -> ModelInput:
        if self.function_name == "generativeai.generate_text":
            return ModelInput(prompt=self.original_kwargs.get("prompt", ""))
        if self.function_name in ["generativeai.chat", "generativeai.chat_async"]:
            prompt = self.original_kwargs.get("messages", [])[-1]
            history = self._extract_history()
            return ModelInput(prompt=prompt, history=history)
        if self.function_name == "generativeai.discuss.ChatResponse.reply":
            prompt = get_argument(
                self.original_args, self.original_kwargs, "message", 1
            )
            history = self._extract_history()
            return ModelInput(prompt=prompt, history=history)

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self, stream: bool, outputs: text_types.Completion | discuss_types.ChatResponse
    ) -> str:
        if stream:
            return self._extract_output_generator(outputs)
        if (
            isinstance(outputs, text_types.Completion)
            and self.function_name == "generativeai.generate_text"
        ):
            return cast(str, outputs.result)
        if isinstance(outputs, discuss_types.ChatResponse) and self.function_name in [
            "generativeai.chat",
            "generativeai.chat_async",
            "generativeai.discuss.ChatResponse.reply",
        ]:
            return cast(str, outputs.messages[-1]["content"])

        raise ValueError(
            f"Unknown function name: {self.function_name} "
            f"or output type: {type(outputs)}"
        )

    def _extract_output_generator(self, outputs: Any) -> str:
        raise NotImplementedError("Google does not support streaming yet.")



# pylint: disable=duplicate-code
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Callable, List, Union, cast

from cohere.client import Client  # type: ignore
from cohere.responses.chat import (  # type: ignore
    Chat,
    StreamEnd,
    StreamingChat,
    StreamStart,
    StreamTextGeneration,
)
from cohere.responses.generation import (  # type: ignore
    Generations,
    StreamingGenerations,
    StreamingText,
)
from typing_extensions import TypeAlias

from nebuly.entities import HistoryEntry, ModelInput
from nebuly.providers.base import PicklerHandler, ProviderDataExtractor
from nebuly.providers.utils import get_argument

logger = logging.getLogger(__name__)


StreamingChatEntry: TypeAlias = Union[StreamStart, StreamEnd, StreamTextGeneration]


def is_cohere_generator(function: Callable[[Any], Any]) -> bool:
    return isinstance(function, (StreamingGenerations, StreamingChat))


class CoherePicklerHandler(PicklerHandler):
    @property
    def _object_key_attribute_mapping(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        names = [
            "api_key",
            "num_workers",
            "request_dict",
            "max_retries",
            "timeout",
            "api_url",
        ]
        return {
            Client: {
                "key_names": names,
                "attribute_names": names,
            },
        }


def handle_cohere_unpickable_objects() -> None:
    pickler_handler = CoherePicklerHandler()
    pickler_handler.handle_unpickable_object(obj=Client)


@dataclass(frozen=True)
class CohereDataExtractor(ProviderDataExtractor):
    original_args: tuple[Any, ...]
    original_kwargs: dict[str, Any]
    function_name: str

    def _extract_history(self) -> list[HistoryEntry]:
        history = get_argument(
            self.original_args, self.original_kwargs, "chat_history", 7
        )

        # Remove messages that are not from the user or the assistant
        history = [
            message
            for message in history
            if message.get("role", message.get("user_name", "")).lower()
            in ["user", "chatbot", "assistant"]
        ]

        if len(history) % 2 != 0:
            logger.warning("Odd number of chat history elements, ignoring last element")
            history = history[:-1]

        # Convert the history to [(user, assistant), ...] format
        history = [
            HistoryEntry(
                user=history[i].get("message", history[i].get("text", "")),
                assistant=history[i + 1].get("message", history[i + 1].get("text", "")),
            )
            for i in range(0, len(history), 2)
            if i < len(history) - 1
        ]
        return history  # type: ignore

    def extract_input_and_history(self) -> ModelInput:
        if self.function_name in ["Client.generate", "AsyncClient.generate"]:
            prompt = get_argument(self.original_args, self.original_kwargs, "prompt", 1)
            return ModelInput(prompt=prompt)
        if self.function_name in ["Client.chat", "AsyncClient.chat"]:
            prompt = get_argument(
                self.original_args, self.original_kwargs, "message", 1
            )
            history = self._extract_history()
            return ModelInput(prompt=prompt, history=history)

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self,
        stream: bool,
        outputs: Generations | Chat | list[StreamingText] | list[StreamingChatEntry],
    ) -> str:
        if stream and isinstance(outputs, list):
            return self._extract_output_generator(outputs)

        if isinstance(outputs, Generations) and self.function_name in [
            "Client.generate",
            "AsyncClient.generate",
        ]:
            return cast(str, outputs.generations[0].text)
        if isinstance(outputs, Chat) and self.function_name in [
            "Client.chat",
            "AsyncClient.chat",
        ]:
            return cast(str, outputs.text)

        raise ValueError(
            f"Unknown function name: {self.function_name} "
            f"or output type: {type(outputs)}"
        )

    def _extract_output_generator(
        self, outputs: list[StreamingText] | list[StreamingChatEntry]
    ) -> str:
        if all(
            isinstance(elem, StreamingText) for elem in outputs
        ) and self.function_name in ["Client.generate", "AsyncClient.generate"]:
            return "".join([output.text for output in outputs])
        if all(
            isinstance(elem, (StreamStart, StreamEnd, StreamTextGeneration))
            for elem in outputs
        ) and self.function_name in ["Client.chat", "AsyncClient.chat"]:
            messages = cast(List[StreamingChatEntry], outputs)
            return "".join(
                [
                    output.text
                    for output in messages
                    if output.event_type == "text-generation"
                ]
            )

        raise ValueError(
            f"Unknown function name: {self.function_name} "
            f"or output type: {type(outputs)}"
        )



from __future__ import annotations

import asyncio
import importlib
import json
import logging
import sys
from copy import deepcopy
from datetime import datetime, timezone
from functools import wraps
from importlib.metadata import version
from inspect import isasyncgenfunction, iscoroutinefunction
from types import ModuleType
from typing import Any, AsyncGenerator, Callable, Generator, Iterable, cast

from packaging.version import parse as parse_version

from nebuly.config import NEBULY_KWARGS
from nebuly.contextmanager import (
    InteractionContext,
    get_nearest_open_interaction,
    new_interaction,
)
from nebuly.entities import HistoryEntry, ModelInput, Observer, Package, SpanWatch
from nebuly.exceptions import NotInInteractionContext
from nebuly.providers.base import ProviderDataExtractor

logger = logging.getLogger(__name__)


def check_no_packages_already_imported(packages: Iterable[Package]) -> None:
    """
    Check that no packages in packages have already been imported.
    """
    for package in packages:
        if package.name in sys.modules:
            logger.debug("%s already imported", package.name)


def import_and_patch_packages(packages: Iterable[Package], observer: Observer) -> None:
    """
    Import each package in packages and patch it with the observer.
    """
    for package in packages:
        try:
            _monkey_patch(package, observer)
        except ImportError:
            pass


def _monkey_patch(package: Package, observer: Observer) -> None:
    module = importlib.import_module(package.name)
    try:
        package_version = version(package.name)
    except Exception:  # pylint: disable=broad-except
        package_version = "unknown"

    if package_version != "unknown":
        max_version_str = package.versions.max_version
        min_version = parse_version(package.versions.min_version)
        pkg_version = parse_version(package_version)

        if max_version_str is None and (
            min_version
            > pkg_version
            != min_version.base_version
            != pkg_version.base_version
        ):
            return
        if (
            max_version_str is not None
            and not min_version <= pkg_version < parse_version(max_version_str)
        ):
            return

    for attr in package.to_patch:
        try:
            _monkey_patch_attribute(attr, module, package_version, observer)
        except (AttributeError, ImportError):
            logger.debug("Failed to patch %s", attr)


def _monkey_patch_attribute(
    attr: str, module: ModuleType, package_version: str, observer: Observer
) -> None:
    tmp_component = module
    path = attr.split(".")
    for i, component_name in enumerate(path[:-1]):
        try:
            tmp_component = getattr(tmp_component, component_name)
        except AttributeError:
            # Handle the case when the module is imported at runtime
            tmp_component = importlib.import_module(
                f"{module.__name__}.{'.'.join(path[:i+1])}"
            )

    setattr(
        tmp_component,
        path[-1],
        _patcher(observer, module.__name__, package_version, attr)(
            getattr(tmp_component, path[-1])
        ),
    )


def _extract_nebuly_kwargs_from_arg(arg: Any, nebuly_kwargs: dict[str, Any]) -> Any:
    if not isinstance(arg, (dict, str)):
        return arg

    needs_json_conversion = False
    if isinstance(arg, str):
        try:
            arg = json.loads(arg)
            if not isinstance(arg, dict):
                # Need this to handle the case when the arg is a string containing
                # for example an int
                return json.dumps(arg)

            needs_json_conversion = True
        except json.JSONDecodeError:
            return arg

    for key in NEBULY_KWARGS:
        if key in arg:
            nebuly_kwargs[key] = arg.pop(key)

    for key, value in arg.items():
        arg[key] = _extract_nebuly_kwargs_from_arg(value, nebuly_kwargs)

    if needs_json_conversion:
        arg = json.dumps(arg)

    return arg


def _split_nebuly_kwargs(
    args: tuple[Any, ...], kwargs: dict[str, Any]
) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Split nebuly kwargs from function kwargs
    """

    nebuly_kwargs = {}
    function_kwargs = {}
    if len(kwargs) > 0:
        for key in kwargs:
            if key in NEBULY_KWARGS:
                nebuly_kwargs[key] = kwargs[key]
            else:
                function_kwargs[key] = kwargs[key]
    else:
        for arg in args:
            arg = _extract_nebuly_kwargs_from_arg(arg, nebuly_kwargs)
    return nebuly_kwargs, function_kwargs


def _get_provider_data_extractor(
    module: str,
    original_args: tuple[Any, ...],
    original_kwargs: dict[str, Any],
    function_name: str,
) -> ProviderDataExtractor:
    constructor: Callable[[Any], Any] | None = None
    if module == "openai":
        if (
            parse_version(version(module)) < parse_version("1.0.0")
            and parse_version(version(module)).base_version != "1.0.0"
        ):
            from nebuly.providers.openai_legacy import (  # pylint: disable=import-outside-toplevel  # noqa: E501
                OpenAILegacyDataExtractor,
            )

            constructor = OpenAILegacyDataExtractor  # type: ignore
        else:
            from nebuly.providers.openai import (  # pylint: disable=import-outside-toplevel  # noqa: E501
                OpenAIDataExtractor,
            )

            constructor = OpenAIDataExtractor  # type: ignore
    if module == "cohere":
        from nebuly.providers.cohere import (  # pylint: disable=import-outside-toplevel
            CohereDataExtractor,
        )

        constructor = CohereDataExtractor  # type: ignore
    if module == "anthropic":
        from nebuly.providers.anthropic import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            AnthropicDataExtractor,
        )

        constructor = AnthropicDataExtractor  # type: ignore

    if module == "google":
        from nebuly.providers.google import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            GoogleDataExtractor,
        )

        constructor = GoogleDataExtractor  # type: ignore

    if module == "transformers":
        from nebuly.providers.huggingface import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            HuggingFaceDataExtractor,
        )

        constructor = HuggingFaceDataExtractor  # type: ignore

    if module == "huggingface_hub":
        from nebuly.providers.huggingface_hub import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            HuggingFaceHubDataExtractor,
        )

        constructor = HuggingFaceHubDataExtractor  # type: ignore

    if module == "vertexai":
        from nebuly.providers.vertexai import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            VertexAIDataExtractor,
        )

        constructor = VertexAIDataExtractor  # type: ignore

    if module == "botocore":
        from nebuly.providers.aws_bedrock import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            AWSBedrockDataExtractor,
        )

        constructor = AWSBedrockDataExtractor  # type: ignore

    if constructor is not None:
        return constructor(  # type: ignore
            original_args=original_args,
            original_kwargs=original_kwargs,
            function_name=function_name,
        )

    raise ValueError(f"Unknown module: {module}")


def _add_span_to_interaction(  # pylint: disable=too-many-arguments
    observer: Observer,
    interaction: InteractionContext,
    user_input: str,
    history: list[HistoryEntry],
    output: str,
    watched: SpanWatch,
    nebuly_kwargs: dict[str, Any],
) -> None:
    interaction._set_observer(observer)  # pylint: disable=protected-access
    interaction._add_span(watched)  # pylint: disable=protected-access
    if interaction.input is None:
        interaction.set_input(user_input)
    if interaction.history is None:
        interaction.set_history(history)
    if interaction.output is None:
        interaction.set_output(output)
    user: str | None = nebuly_kwargs.get("user_id")
    if interaction.user is None and user is not None:
        interaction._set_user(user)  # pylint: disable=protected-access
    user_group_profile: str | None = nebuly_kwargs.get("user_group_profile")
    if interaction.user_group_profile is None and user_group_profile is not None:
        interaction._set_user_group_profile(  # pylint: disable=protected-access
            user_group_profile
        )


def _add_interaction_span(  # pylint: disable=too-many-arguments, too-many-locals
    original_args: tuple[Any, ...],
    original_kwargs: dict[str, Any],
    module: str,
    function_name: str,
    output: Any,
    observer: Observer,
    watched: SpanWatch,
    nebuly_kwargs: dict[str, Any],
    start_time: datetime,
    stream: bool = False,
) -> None:
    try:
        provider_data_extractor = _get_provider_data_extractor(
            module, original_args, original_kwargs, function_name
        )
        model_input_res = provider_data_extractor.extract_input_and_history()
        model_output_res = provider_data_extractor.extract_output(stream, output)
    except ValueError:
        logger.debug("Unknown module: %s", function_name)
        model_input_res = ModelInput(prompt="")
        model_output_res = ""

    try:
        interaction = get_nearest_open_interaction()
        if isinstance(model_input_res, list):
            logger.warning(
                "Batch prediction found inside a user defined interaction, only the "
                "first element will be used"
            )
            model_input_res = model_input_res[0]
            model_output_res = model_output_res[0]

        _add_span_to_interaction(
            observer=observer,
            interaction=interaction,
            user_input=model_input_res.prompt,
            history=model_input_res.history,
            output=cast(str, model_output_res),
            watched=watched,
            nebuly_kwargs=nebuly_kwargs,
        )
    except NotInInteractionContext:
        if isinstance(model_input_res, list):
            for model_input, model_output in zip(model_input_res, model_output_res):
                with new_interaction() as interaction:
                    interaction.time_start = start_time
                    _add_span_to_interaction(
                        observer=observer,
                        interaction=interaction,
                        user_input=model_input.prompt,
                        history=model_input.history,
                        output=model_output,
                        watched=watched,
                        nebuly_kwargs=nebuly_kwargs,
                    )
        else:
            with new_interaction() as interaction:
                interaction.time_start = start_time
                _add_span_to_interaction(
                    observer=observer,
                    interaction=interaction,
                    user_input=model_input_res.prompt,
                    history=model_input_res.history,
                    output=cast(str, model_output_res),
                    watched=watched,
                    nebuly_kwargs=nebuly_kwargs,
                )


def watch_from_generator(  # pylint: disable=too-many-arguments
    *,
    generator: Generator[Any, Any, Any],
    observer: Observer,
    module: str,
    package_version: str,
    function_name: str,
    called_start: datetime,
    original_args: tuple[Any, ...],
    original_kwargs: dict[str, Any],
    nebuly_kwargs: dict[str, Any],
) -> Generator[Any, Any, Any]:
    """
    Watch a generator

    Creates the SpanWatch object while the generator is being iterated over.
    Waits until the iteration is done to call the observer.
    """
    original_result = []
    generator_first_element_timestamp = None

    first_element = True
    for element in generator:
        if first_element:
            first_element = False
            generator_first_element_timestamp = datetime.now(timezone.utc)
        logger.debug("Yielding %s", element)
        original_result.append(deepcopy(element))
        yield element

    called_end = datetime.now(timezone.utc)

    watched = SpanWatch(
        module=module,
        version=package_version,
        function=function_name,
        called_start=called_start,
        called_end=called_end,
        called_with_args=original_args,
        called_with_kwargs=original_kwargs,
        returned=original_result,
        generator=True,
        generator_first_element_timestamp=generator_first_element_timestamp,
        provider_extras=nebuly_kwargs,
    )

    _add_interaction_span(
        original_args=original_args,
        original_kwargs=original_kwargs,
        module=module,
        function_name=function_name,
        output=original_result,
        observer=observer,
        watched=watched,
        nebuly_kwargs=nebuly_kwargs,
        stream=True,
        start_time=called_start,
    )


async def watch_from_generator_async(  # pylint: disable=too-many-arguments
    *,
    generator: AsyncGenerator[Any, Any],
    observer: Observer,
    module: str,
    package_version: str,
    function_name: str,
    called_start: datetime,
    original_args: tuple[Any, ...],
    original_kwargs: dict[str, Any],
    nebuly_kwargs: dict[str, Any],
) -> AsyncGenerator[Any, Any]:
    """
    Watch a generator

    Creates the SpanWatch object while the generator is being iterated over.
    Waits until the iteration is done to call the observer.
    """
    original_result = []
    generator_first_element_timestamp = None

    first_element = True
    async for element in generator:
        if first_element:
            first_element = False
            generator_first_element_timestamp = datetime.now(timezone.utc)
        logger.debug("Yielding %s", element)
        original_result.append(deepcopy(element))
        yield element

    called_end = datetime.now(timezone.utc)

    watched = SpanWatch(
        module=module,
        version=package_version,
        function=function_name,
        called_start=called_start,
        called_end=called_end,
        called_with_args=original_args,
        called_with_kwargs=original_kwargs,
        returned=original_result,
        generator=True,
        generator_first_element_timestamp=generator_first_element_timestamp,
        provider_extras=nebuly_kwargs,
    )
    _add_interaction_span(
        original_args=original_args,
        original_kwargs=original_kwargs,
        module=module,
        function_name=function_name,
        output=original_result,
        observer=observer,
        watched=watched,
        nebuly_kwargs=nebuly_kwargs,
        stream=True,
        start_time=called_start,
    )


def _handle_unpickleable_objects(module: str, args: Any) -> None:
    """
    Make unpickleable objects pickleable to avoid errors when
    using the deepcopy function
    """
    try:
        from nebuly.providers.openai import (  # pylint: disable=import-outside-toplevel
            handle_openai_unpickable_objects,
        )

        handle_openai_unpickable_objects()
    except ImportError:
        pass
    try:
        from nebuly.providers.cohere import (  # pylint: disable=import-outside-toplevel
            handle_cohere_unpickable_objects,
        )

        handle_cohere_unpickable_objects()
    except ImportError:
        pass

    try:
        from nebuly.providers.anthropic import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            handle_anthropic_unpickable_objects,
        )

        handle_anthropic_unpickable_objects()
    except ImportError:
        pass

    try:
        from nebuly.providers.google import (  # pylint: disable=import-outside-toplevel
            handle_google_unpickable_objects,
        )

        handle_google_unpickable_objects()
    except ImportError:
        pass

    try:
        from nebuly.providers.vertexai import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            handle_vertexai_unpickable_objects,
        )

        handle_vertexai_unpickable_objects()
    except ImportError:
        pass
    if module == "botocore":
        try:
            from nebuly.providers.aws_bedrock import (  # pylint: disable=import-outside-toplevel  # noqa: E501
                handle_aws_bedrock_unpickable_objects,
            )

            handle_aws_bedrock_unpickable_objects(args[0])
        except ImportError:
            pass


def _setup_args_kwargs(
    *args: Any, **kwargs: Any
) -> tuple[tuple[Any, ...], dict[str, Any], dict[str, Any], dict[str, Any]]:
    """
    Split nebuly kwargs from function kwargs

    Returns:
        original_args: deepcopy of args passed to the function
        original_kwargs: deepcopy of kwargs passed to the function that are not
            nebuly kwargs
        function_kwargs: kwargs passed to the function that are not nebuly kwargs
        nebuly_kwargs: kwargs passed to the function that are nebuly kwargs
    """
    nebuly_kwargs, function_kwargs = _split_nebuly_kwargs(args, kwargs)
    original_args = deepcopy(args)
    nebuly_kwargs = deepcopy(nebuly_kwargs)
    original_kwargs = deepcopy(function_kwargs)

    return original_args, original_kwargs, function_kwargs, nebuly_kwargs


def _is_generator(obj: Any, module: str) -> bool:
    if isinstance(obj, (Generator, AsyncGenerator)):
        return True

    try:
        from nebuly.providers.openai import (  # pylint: disable=import-outside-toplevel
            is_openai_generator,
        )

        if is_openai_generator(obj):
            return True
    except ImportError:
        pass

    try:
        from nebuly.providers.cohere import (  # pylint: disable=import-outside-toplevel
            is_cohere_generator,
        )

        if is_cohere_generator(obj):
            return True
    except ImportError:
        pass

    try:
        from nebuly.providers.anthropic import (  # pylint: disable=import-outside-toplevel  # noqa: E501
            is_anthropic_generator,
        )

        if is_anthropic_generator(obj):
            return True
    except ImportError:
        pass

    if module == "botocore":
        try:
            from nebuly.providers.aws_bedrock import (  # pylint: disable=import-outside-toplevel  # noqa: E501
                is_aws_bedrock_generator,
            )

            if is_aws_bedrock_generator(obj):
                return True
        except ImportError:
            pass

    return False


def coroutine_wrapper(
    f: Callable[[Any], Any],
    observer: Observer,
    module: str,
    package_version: str,
    function_name: str,
) -> Callable[[Any], Any]:
    @wraps(f)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        logger.debug("Calling %s.%s", module, function_name)

        if module == "langchain":
            from nebuly.providers.langchain import (  # pylint: disable=import-outside-toplevel  # noqa: E501
                wrap_langchain_async,
            )

            return wrap_langchain_async(
                observer=observer,
                function_name=function_name,
                f=f,
                args=args,
                kwargs=kwargs,
            )

        _handle_unpickleable_objects(module, args)

        (
            original_args,
            original_kwargs,
            function_kwargs,
            nebuly_kwargs,
        ) = _setup_args_kwargs(*args, **kwargs)

        generator_first_element_timestamp = None

        called_start = datetime.now(timezone.utc)

        if isasyncgenfunction(f):
            result = f(*args, **function_kwargs)
        else:
            result = await f(*args, **function_kwargs)

        if _is_generator(result, module):
            logger.debug("Result is a generator")
            return watch_from_generator_async(
                generator=result,
                observer=observer,
                module=module,
                package_version=package_version,
                function_name=function_name,
                called_start=called_start,
                original_args=original_args,
                original_kwargs=original_kwargs,
                nebuly_kwargs=nebuly_kwargs,
            )

        logger.debug("Result is not a generator")

        original_result = deepcopy(result)
        called_end = datetime.now(timezone.utc)
        watched = SpanWatch(
            module=module,
            version=package_version,
            function=function_name,
            called_start=called_start,
            called_end=called_end,
            called_with_args=original_args,
            called_with_kwargs=original_kwargs,
            returned=original_result,
            generator=False,
            generator_first_element_timestamp=generator_first_element_timestamp,
            provider_extras=nebuly_kwargs,
        )
        _add_interaction_span(
            original_args=original_args,
            original_kwargs=original_kwargs,
            module=module,
            function_name=function_name,
            output=original_result,
            observer=observer,
            watched=watched,
            nebuly_kwargs=nebuly_kwargs,
            start_time=called_start,
        )
        return result

    return wrapper


def function_wrapper(
    f: Callable[[Any], Any],
    observer: Observer,
    module: str,
    package_version: str,
    function_name: str,
) -> Callable[[Any], Any]:
    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        logger.debug("Calling %s.%s", module, function_name)

        if module == "langchain":
            from nebuly.providers.langchain import (  # pylint: disable=import-outside-toplevel  # noqa: E501
                wrap_langchain,
            )

            return wrap_langchain(
                observer=observer,
                function_name=function_name,
                f=f,
                args=args,
                kwargs=kwargs,
            )
        if module == "botocore":
            from nebuly.providers.aws_bedrock import (  # pylint: disable=import-outside-toplevel  # noqa: E501
                is_model_supported,
            )

            if args[1] not in [
                "InvokeModel",
                "InvokeModelWithResponseStream",
            ] or not is_model_supported(args[2]["modelId"]):
                for key in NEBULY_KWARGS:
                    if key in args[2]:
                        args[2].pop(key)
                return f(*args, **kwargs)

        _handle_unpickleable_objects(module, args)

        (
            original_args,
            original_kwargs,
            function_kwargs,
            nebuly_kwargs,
        ) = _setup_args_kwargs(*args, **kwargs)

        generator_first_element_timestamp = None

        called_start = datetime.now(timezone.utc)
        result = f(*args, **function_kwargs)

        if _is_generator(result, module):
            logger.debug("Result is a generator")
            if module == "botocore":
                # AWS case must be handled separatly because has a different return type
                result["body"] = watch_from_generator(
                    generator=result["body"],
                    observer=observer,
                    module=module,
                    package_version=package_version,
                    function_name=function_name,
                    called_start=called_start,
                    original_args=original_args,
                    original_kwargs=original_kwargs,
                    nebuly_kwargs=nebuly_kwargs,
                )
                return result
            return watch_from_generator(
                generator=result,
                observer=observer,
                module=module,
                package_version=package_version,
                function_name=function_name,
                called_start=called_start,
                original_args=original_args,
                original_kwargs=original_kwargs,
                nebuly_kwargs=nebuly_kwargs,
            )

        logger.debug("Result is not a generator")

        original_result = deepcopy(result)
        called_end = datetime.now(timezone.utc)
        watched = SpanWatch(
            module=module,
            version=package_version,
            function=function_name,
            called_start=called_start,
            called_end=called_end,
            called_with_args=original_args,
            called_with_kwargs=original_kwargs,
            returned=original_result,
            generator=False,
            generator_first_element_timestamp=generator_first_element_timestamp,
            provider_extras=nebuly_kwargs,
        )
        _add_interaction_span(
            original_args=original_args,
            original_kwargs=original_kwargs,
            module=module,
            function_name=function_name,
            output=original_result,
            observer=observer,
            watched=watched,
            nebuly_kwargs=nebuly_kwargs,
            start_time=called_start,
        )
        return result

    return wrapper


def _patcher(
    observer: Observer, module: str, package_version: str, function_name: str
) -> Callable[[Any], Any]:
    """
    Decorator that calls observer with a SpanWatch instance when the decorated
    function is called

    kwargs that start with nebuly_ are passed to the observer and not the
    decorated function
    """

    def inner(f: Callable[[Any], Any]) -> Callable[[Any], Any]:
        if (
            iscoroutinefunction(f)
            or asyncio.iscoroutinefunction(f)  # Needed for python 3.9
            or isasyncgenfunction(f)
            or "async" in function_name.lower()
        ):
            return coroutine_wrapper(
                f, observer, module, package_version, function_name
            )

        return function_wrapper(f, observer, module, package_version, function_name)

    return inner



from __future__ import annotations

import inspect
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Dict, Generator, List, Tuple, cast
from uuid import UUID

from nebuly.entities import (
    EventType,
    HistoryEntry,
    InteractionWatch,
    Observer,
    SpanWatch,
)
from nebuly.exceptions import (
    AlreadyInInteractionContext,
    InteractionContextInitiationError,
    InteractionMustBeLocalVariable,
    MissingRequiredNebulyFieldError,
    NotInInteractionContext,
)


@dataclass
class EventData:
    type: EventType
    args: tuple[Any, ...] | None = None
    kwargs: dict[str, Any] | None = None
    output: Any | None = None

    def add_end_event_data(
        self,
        output: Any,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> None:
        if kwargs is not None:
            kwargs.pop("parent_run_id")
        if self.args is None:
            self.args = tuple()
        if self.kwargs is None:
            self.kwargs = {}
        self.args += args if args is not None else tuple()
        self.kwargs = dict(self.kwargs, **kwargs) if kwargs is not None else self.kwargs
        self.output = output


@dataclass
class EventHierarchy:
    parent_run_id: UUID
    root_run_id: UUID


@dataclass
class Event:
    event_id: UUID
    hierarchy: EventHierarchy | None
    data: EventData
    module: str
    start_time: datetime
    end_time: datetime | None = None

    def as_span_watch(self) -> SpanWatch:
        if self.end_time is None:
            raise ValueError("Event has not been finished yet.")
        return SpanWatch(
            span_id=self.event_id,
            module=self.module,
            version="unknown",
            function=self._get_function(),
            called_start=self.start_time,
            called_end=self.end_time,
            called_with_args=cast(Tuple[Any], self.data.args),
            called_with_kwargs=cast(Dict[str, Any], self.data.kwargs),
            returned=self.data.output,
            generator=False,
            generator_first_element_timestamp=None,
            provider_extras={
                "event_type": self.data.type.value,
            },
            rag_source=self._get_rag_source(),
        )

    def _get_function(self) -> str:
        if self.data.kwargs is None or len(self.data.kwargs) == 0:
            raise ValueError("Event has no kwargs.")
        if self.data.type is EventType.TOOL:
            return self.data.kwargs["serialized"]["name"]  # type: ignore
        return ".".join(self.data.kwargs["serialized"]["id"])

    def _get_rag_source(self) -> str | None:
        if self.data.kwargs is None or len(self.data.kwargs) == 0:
            raise ValueError("Event has no kwargs.")
        if self.data.type is EventType.TOOL:
            return self.data.kwargs["serialized"]["name"]  # type: ignore
        if self.data.type is EventType.RETRIEVAL:
            return self.data.kwargs["serialized"]["id"][-1]  # type: ignore

        return None

    def set_end_time(self) -> None:
        self.end_time = datetime.now(timezone.utc)


@dataclass
class EventsStorage:
    events: dict[UUID, Event] = field(default_factory=dict)

    def get_root_id(self, event_id: UUID) -> UUID:
        if event_id not in self.events:
            raise ValueError(f"Event {event_id} not found in events events storage.")
        event = self.events[event_id]
        if event.hierarchy is None:
            return event_id
        return self.get_root_id(event.hierarchy.root_run_id)

    def add_event(
        self,
        event_id: UUID,
        parent_id: UUID | None,
        data: EventData,
        module: str,
    ) -> None:
        if event_id in self.events:
            raise ValueError(f"Event {event_id} already exists in events storage.")

        # Handle new event
        if parent_id is None:
            hierarchy = None
        else:
            hierarchy = EventHierarchy(
                parent_run_id=parent_id, root_run_id=self.get_root_id(parent_id)
            )
        self.events[event_id] = Event(
            event_id=event_id,
            hierarchy=hierarchy,
            data=data,
            module=module,
            start_time=datetime.now(timezone.utc),
        )

    def delete_events(self, root_id: UUID) -> None:
        if root_id not in self.events:
            raise ValueError(f"Event {root_id} not found in events hierarchy storage.")

        keys_to_delete = [root_id]

        for event_id, event_detail in self.events.items():
            if (
                event_detail.hierarchy is not None
                and event_detail.hierarchy.root_run_id == root_id
            ):
                keys_to_delete.append(event_id)

        for key in keys_to_delete:
            self.events.pop(key)

    def get_spans(self) -> list[SpanWatch]:
        return [event.as_span_watch() for event in self.events.values()]


class InteractionContext:  # pylint: disable=too-many-instance-attributes
    def __init__(  # pylint: disable=too-many-arguments
        self,
        user: str | None = None,
        user_group_profile: str | None = None,
        initial_input: str | None = None,
        final_output: str | None = None,
        history: list[HistoryEntry] | None = None,
        hierarchy: dict[UUID, UUID | None] | None = None,
        spans: list[SpanWatch] | None = None,
        do_not_call_directly: bool = False,
    ) -> None:
        if not do_not_call_directly:
            raise InteractionContextInitiationError(
                "Interaction cannot be directly instantiate, use the"
                " 'new_interaction' contextmanager"
            )
        self._events_storage: EventsStorage = EventsStorage()
        self._observer: Observer | None = None
        self._finished: bool = False

        self.user = user
        self.user_group_profile = user_group_profile
        self.input = initial_input
        self.output = final_output
        self.spans = [] if spans is None else spans
        self.history = history
        self.hierarchy = hierarchy
        self.time_start = datetime.now(timezone.utc)

    def set_input(self, value: str) -> None:
        self.input = value

    def set_history(self, value: list[dict[str, str]] | list[HistoryEntry]) -> None:
        if len(value) == 0:
            self.history = []
        elif isinstance(value[0], dict):
            value = cast(List[Dict[str, str]], value)
            history = [
                message
                for message in value
                if message["role"].lower() in ["user", "assistant"]
            ]
            if len(history) % 2 != 0:
                raise ValueError(
                    "Odd number of chat history elements, please provide "
                    "a valid history."
                )
            self.history = [
                HistoryEntry(
                    user=history[i]["content"], assistant=history[i + 1]["content"]
                )
                for i in range(len(history) - 1)
            ]
        else:
            value = cast(List[HistoryEntry], value)
            self.history = value

    def set_output(self, value: str) -> None:
        self.output = value

    def _set_observer(self, observer: Observer) -> None:
        self._observer = observer

    def _add_span(self, value: SpanWatch) -> None:
        self.spans.append(value)

    def _finish(self) -> None:
        self._finished = True
        if len(self._events_storage.events) > 0:
            self.spans += self._events_storage.get_spans()
            self.hierarchy = {
                event.event_id: event.hierarchy.parent_run_id
                if event.hierarchy is not None
                else None
                for event in self._events_storage.events.values()
            }
        else:
            self.hierarchy = {}
        for span in self.spans:
            if span.provider_extras is None:
                continue
            parent_id: UUID | None = span.provider_extras.get("parent_run_id")
            if parent_id is not None:
                self.hierarchy[span.span_id] = parent_id
        if self._observer is not None:
            self._observer(self._as_interaction_watch())

    def _set_user(self, value: str) -> None:
        self.user = value

    def _set_user_group_profile(self, value: str | None) -> None:
        self.user_group_profile = value

    def _validate_interaction(self) -> None:
        if self.input is None:
            raise ValueError("Interaction has no input.")
        if self.output is None:
            raise ValueError("Interaction has no output.")
        if self.history is None:
            raise ValueError("Interaction has no history.")
        if self.hierarchy is None:
            raise ValueError("Interaction has no hierarchy.")
        if self.user is None:
            raise MissingRequiredNebulyFieldError(
                "Missing required nebuly field: 'user_id'. Please add it when calling"
                "the original provider method."
            )

    def _as_interaction_watch(self) -> InteractionWatch:
        self._validate_interaction()
        return InteractionWatch(
            input=self.input,  # type: ignore
            output=self.output,  # type: ignore
            time_start=self.time_start,
            time_end=datetime.now(timezone.utc),
            spans=self.spans,
            history=self.history,  # type: ignore
            hierarchy=self.hierarchy,  # type: ignore
            end_user=self.user,  # type: ignore
            end_user_group_profile=self.user_group_profile,
        )


def get_nearest_open_interaction() -> InteractionContext:
    frames = inspect.stack()
    for frame in frames[::-1]:
        for v in frame.frame.f_locals.values():
            if (
                isinstance(v, InteractionContext)
                and not v._finished  # pylint: disable=protected-access
            ):
                return v
    raise NotInInteractionContext()


@contextmanager
def new_interaction(
    user_id: str | None = None, user_group_profile: str | None = None
) -> Generator[InteractionContext, None, None]:
    try:
        get_nearest_open_interaction()
        raise AlreadyInInteractionContext()
    except NotInInteractionContext:
        pass

    try:
        yield InteractionContext(user_id, user_group_profile, do_not_call_directly=True)
    finally:
        try:
            interaction = get_nearest_open_interaction()
        except NotInInteractionContext:
            raise InteractionMustBeLocalVariable()  # pylint: disable=raise-missing-from
        else:
            interaction._finish()  # pylint: disable=protected-access



from __future__ import annotations

from nebuly.entities import InteractionWatch, Publisher


class NebulyObserver:
    """
    NebulyObserver is an observer that sends a message to the API when a
    patched function is called.
    """

    def __init__(
        self,
        *,
        api_key: str,
        publish: Publisher,
    ) -> None:
        self._api_key = api_key
        self._publisher = publish

    def on_event_received(self, interaction_watch: InteractionWatch) -> None:
        self._publisher(interaction_watch)



from __future__ import annotations

from nebuly.entities import Package, SupportedVersion

NEBULY_KWARGS = {
    "user_id",
    "user_group_profile",
    "parent_run_id",
    "root_run_id",
}

PACKAGES = (
    Package(
        "openai",
        SupportedVersion("0.10.0", "0.30.0"),
        (
            "Completion.create",
            "Completion.acreate",
            "ChatCompletion.create",
            "ChatCompletion.acreate",
        ),
    ),
    Package(
        "openai",
        SupportedVersion("1.0.0"),
        (
            "resources.chat.completions.Completions.create",
            "resources.chat.completions.AsyncCompletions.create",
            "resources.completions.Completions.create",
            "resources.completions.AsyncCompletions.create",
        ),
    ),
    Package(
        "langchain",
        SupportedVersion("0.0.200"),
        (
            "chains.base.Chain.__call__",
            "chains.base.Chain.acall",
            "llms.base.BaseLLM.generate",
            "llms.base.BaseLLM.agenerate",
            "chat_models.base.BaseChatModel.generate",
            "chat_models.base.BaseChatModel.agenerate",
            "indexes.vectorstore.VectorStoreIndexWrapper.query",
        ),
    ),
    Package(
        "cohere",
        SupportedVersion("4.0.0"),
        (
            "Client.generate",
            "AsyncClient.generate",
            "Client.chat",
            "AsyncClient.chat",
        ),
    ),
    Package(
        "anthropic",
        SupportedVersion("0.3.0"),
        (
            "resources.Completions.create",
            "resources.AsyncCompletions.create",
        ),
    ),
    Package(
        "huggingface_hub",
        SupportedVersion("0.12.0"),
        (
            "InferenceClient.conversational",
            "InferenceClient.text_generation",
            "AsyncInferenceClient.conversational",
            "AsyncInferenceClient.text_generation",
        ),
    ),
    Package(
        "transformers",
        SupportedVersion("4.10.0"),
        (
            "pipelines.conversational.ConversationalPipeline.__call__",
            "pipelines.text_generation.TextGenerationPipeline.__call__",
        ),
    ),
    Package(
        "google",
        SupportedVersion("0.0.1"),
        (
            "generativeai.generate_text",
            "generativeai.chat",
            "generativeai.chat_async",
            "generativeai.discuss.ChatResponse.reply",
        ),
    ),
    Package(
        "vertexai",
        SupportedVersion("0.0.1"),
        (
            "language_models.TextGenerationModel.predict",
            "language_models.TextGenerationModel.predict_async",
            "language_models.TextGenerationModel.predict_streaming",
            "language_models.ChatSession.send_message",
            "language_models.ChatSession.send_message_async",
            "language_models.ChatSession.send_message_streaming",
        ),
    ),
    Package(
        "botocore",
        SupportedVersion("1.30.0"),
        ("client.BaseClient._make_api_call",),
    ),
)



# pylint: disable=duplicate-code
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Callable, cast

from vertexai.language_models import (  # type: ignore
    ChatMessage,
    ChatModel,
    ChatSession,
    TextGenerationModel,
    TextGenerationResponse,
)

from nebuly.entities import HistoryEntry, ModelInput
from nebuly.providers.base import PicklerHandler, ProviderDataExtractor
from nebuly.providers.utils import get_argument

logger = logging.getLogger(__name__)


class VertexAIPicklerHandler(PicklerHandler):
    @property
    def _object_key_attribute_mapping(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        return {
            TextGenerationModel: {
                "key_names": ["model_id", "endpoint_name"],
                "attribute_names": ["_model_id", "_endpoint_name"],
            },
            ChatModel: {
                "key_names": ["model_id", "endpoint_name"],
                "attribute_names": ["_model_id", "_endpoint_name"],
            },
            ChatSession: {
                "key_names": [
                    "model",
                    "context",
                    "examples",
                    "max_output_tokens",
                    "temperature",
                    "top_k",
                    "top_p",
                    "message_history",
                    "stop_sequences",
                ],
                "attribute_names": [
                    "_model._model_id",
                    "_context",
                    "_examples",
                    "_max_output_tokens",
                    "_temperature",
                    "_top_k",
                    "_top_p",
                    "_message_history",
                    "_stop_sequences",
                ],
            },
        }


def handle_vertexai_unpickable_objects() -> None:
    pickler_handler = VertexAIPicklerHandler()

    for obj in [TextGenerationModel, ChatModel, ChatSession]:
        pickler_handler.handle_unpickable_object(
            obj=obj,
        )


@dataclass(frozen=True)
class VertexAIDataExtractor(ProviderDataExtractor):
    original_args: tuple[Any, ...]
    original_kwargs: dict[str, Any]
    function_name: str

    def _extract_history(self) -> list[HistoryEntry]:
        message_history: list[ChatMessage] = getattr(
            self.original_args[0], "message_history", []
        )

        # Remove messages that are not from the user or the assistant
        message_history = [
            message
            for message in message_history
            if len(message_history) > 1 and message.author.lower() in ["user", "bot"]
        ]

        if len(message_history) % 2 != 0:
            logger.warning("Odd number of chat history elements, ignoring last element")
            message_history = message_history[:-1]

        # Convert the history to [(user, assistant), ...] format
        history: list[HistoryEntry] = [
            HistoryEntry(
                user=message_history[i].content,
                assistant=message_history[i + 1].content,
            )
            for i in range(0, len(message_history), 2)
            if i < len(message_history) - 1
        ]

        return history

    def extract_input_and_history(self) -> ModelInput:
        if self.function_name in [
            "language_models.ChatSession.send_message",
            "language_models.ChatSession.send_message_async",
            "language_models.ChatSession.send_message_streaming",
            "language_models.TextGenerationModel.predict",
            "language_models.TextGenerationModel.predict_async",
            "language_models.TextGenerationModel.predict_streaming",
        ]:
            prompt = get_argument(
                args=self.original_args,
                kwargs=self.original_kwargs,
                arg_name="prompt"
                if "TextGenerationModel" in self.function_name
                else "message",
                arg_idx=1,
            )
            history = self._extract_history()
            return ModelInput(prompt=prompt, history=history)

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self,
        stream: bool,
        outputs: TextGenerationResponse | list[TextGenerationResponse],
    ) -> str:
        if stream and isinstance(outputs, list):
            return self._extract_output_generator(outputs)
        if isinstance(outputs, TextGenerationResponse) and self.function_name in [
            "language_models.TextGenerationModel.predict",
            "language_models.TextGenerationModel.predict_async",
            "language_models.ChatSession.send_message",
            "language_models.ChatSession.send_message_async",
        ]:
            return cast(str, outputs.text)

        raise ValueError(
            f"Unknown function name: {self.function_name} "
            f"or output type: {type(outputs)}"
        )

    def _extract_output_generator(self, outputs: list[TextGenerationResponse]) -> str:
        if all(
            isinstance(output, TextGenerationResponse) for output in outputs
        ) and self.function_name in [
            "language_models.TextGenerationModel.predict_streaming",
            "language_models.ChatSession.send_message_streaming",
        ]:
            return "".join([output.text for output in outputs])

        raise ValueError(
            f"Unknown function name: {self.function_name} "
            f"or output type: {type(outputs)}"
        )



# pylint: disable=duplicate-code, import-error, no-name-in-module
# mypy: ignore-errors
from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from typing import Any, Callable, cast

import openai
from openai import AsyncOpenAI, OpenAI, _ModuleClient  # type: ignore  # noqa: E501
from openai.types.chat.chat_completion import (  # type: ignore  # noqa: E501
    ChatCompletion,
    Choice,
)
from openai.types.chat.chat_completion_chunk import (  # type: ignore  # noqa: E501
    ChatCompletionChunk,
)
from openai.types.completion import Completion  # type: ignore  # noqa: E501
from openai.types.completion_choice import (  # type: ignore  # noqa: E501
    CompletionChoice,
)

from nebuly.entities import HistoryEntry, ModelInput
from nebuly.providers.base import PicklerHandler, ProviderDataExtractor

logger = logging.getLogger(__name__)


class OpenAIPicklerHandler(PicklerHandler):
    @property
    def _object_key_attribute_mapping(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        names = ["api_key", "organization"]
        return {
            key: {
                "key_names": names,
                "attribute_names": names,
            }
            for key in [OpenAI, AsyncOpenAI, _ModuleClient]
        }


def handle_openai_unpickable_objects() -> None:
    pickler_handler = OpenAIPicklerHandler()
    for obj in [OpenAI, AsyncOpenAI, _ModuleClient]:
        pickler_handler.handle_unpickable_object(
            obj=obj,
        )


def is_openai_generator(obj: Any) -> bool:
    return isinstance(
        obj, (openai.Stream, openai.AsyncStream)  # pylint: disable=no-member
    )


@dataclass(frozen=True)
class OpenAIDataExtractor(ProviderDataExtractor):
    original_args: tuple[Any, ...]
    original_kwargs: dict[str, Any]
    function_name: str

    def _extract_history(self) -> list[HistoryEntry]:
        history = self.original_kwargs.get("messages", [])[:-1]

        # Remove messages that are not from the user or the assistant
        history = [
            message
            for message in history
            if len(history) > 1 and message["role"].lower() in ["user", "assistant"]
        ]

        if len(history) % 2 != 0:
            logger.warning("Odd number of chat history elements, ignoring last element")
            history = history[:-1]

        # Convert the history to [(user, assistant), ...] format
        history = [
            HistoryEntry(
                user=history[i]["content"], assistant=history[i + 1]["content"]
            )
            for i in range(0, len(history), 2)
            if i < len(history) - 1
        ]
        return history

    def extract_input_and_history(self) -> ModelInput:
        if self.function_name in [
            "resources.completions.Completions.create",
            "resources.completions.AsyncCompletions.create",
        ]:
            return ModelInput(prompt=self.original_kwargs.get("prompt", ""))
        if self.function_name in [
            "resources.chat.completions.Completions.create",
            "resources.chat.completions.AsyncCompletions.create",
        ]:
            prompt = self.original_kwargs.get("messages", [])[-1]["content"]
            history = self._extract_history()
            return ModelInput(prompt=prompt, history=history)

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self,
        stream: bool,
        outputs: ChatCompletion
        | Completion
        | list[ChatCompletionChunk]
        | list[Completion],
    ) -> str:
        if isinstance(outputs, list) and stream:
            return self._extract_output_generator(outputs)

        if isinstance(outputs, Completion) and self.function_name in [
            "resources.completions.Completions.create",
            "resources.completions.AsyncCompletions.create",
        ]:
            return cast(CompletionChoice, outputs.choices[0]).text
        if isinstance(outputs, ChatCompletion) and self.function_name in [
            "resources.chat.completions.Completions.create",
            "resources.chat.completions.AsyncCompletions.create",
        ]:
            if cast(Choice, outputs.choices[0]).message.content is not None:
                # Normal chat completion
                return cast(str, cast(Choice, outputs.choices[0]).message.content)
            # Chat completion with function call
            return json.dumps(
                {
                    "function_name": cast(
                        Choice, outputs.choices[0]
                    ).message.function_call.name,
                    "arguments": cast(
                        Choice, outputs.choices[0]
                    ).message.function_call.arguments,
                }
            )

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"output type: {type(outputs)}"
        )

    def _extract_output_generator(
        self, outputs: list[Completion] | list[ChatCompletionChunk]
    ) -> str:
        if all(
            isinstance(output, Completion) for output in outputs
        ) and self.function_name in [
            "resources.completions.Completions.create",
            "resources.completions.AsyncCompletions.create",
        ]:
            return "".join(
                [getattr(output.choices[0], "text", "") or "" for output in outputs]
            )
        if all(
            isinstance(output, ChatCompletionChunk) for output in outputs
        ) and self.function_name in [
            "resources.chat.completions.Completions.create",
            "resources.chat.completions.AsyncCompletions.create",
        ]:
            if not all(
                getattr(output.choices[0].delta, "content") is None
                for output in outputs
            ):
                # Normal chat completion
                return "".join(
                    [
                        getattr(output.choices[0].delta, "content", "") or ""
                        for output in outputs
                    ]
                )
            # Chat completion with function call
            return json.dumps(
                {
                    "function_name": "".join(
                        getattr(output.choices[0].delta.function_call, "name", "") or ""
                        for output in outputs
                    ),
                    "arguments": "".join(
                        getattr(output.choices[0].delta.function_call, "arguments", "")
                        or ""
                        for output in outputs
                    ),
                }
            )

        raise ValueError(
            f"Unknown function name: {self.function_name} "
            f"or output type: {type(outputs)}"
        )



from __future__ import annotations

import logging
from inspect import isasyncgenfunction
from typing import Any, Callable, cast
from uuid import UUID

from langchain.callbacks.manager import CallbackManager
from langchain.chains.base import Chain
from langchain.prompts import ChatPromptTemplate, PromptTemplate
from langchain.prompts.chat import AIMessagePromptTemplate, HumanMessagePromptTemplate

from nebuly.contextmanager import get_nearest_open_interaction, new_interaction
from nebuly.entities import HistoryEntry, ModelInput, Observer
from nebuly.exceptions import NotInInteractionContext
from nebuly.providers.utils import get_argument
from nebuly.tracking_handlers import LangChainTrackingHandler

logger = logging.getLogger(__name__)


def _get_tracking_info_for_provider_call(**kwargs: Any) -> dict[str, Any]:
    """
    This function is a hack to add the chain run_ids info to the kwargs of the
    provider call. This is needed to associate each call to a provider (ex OpenAI)
    with the belonging chain.
    """
    callbacks = kwargs.get("callbacks")
    if not isinstance(callbacks, CallbackManager) or callbacks.parent_run_id is None:
        # If the llm/chat_model is called without a CallbackManager, it's not
        # called from a chain, so we don't need to add the run ids info
        return {}

    # Get the parent_run_id from the CallbackManager
    callback_manager = callbacks
    parent_run_id = cast(UUID, callback_manager.parent_run_id)
    additional_kwargs = {"parent_run_id": str(parent_run_id)}

    # Get the root_run_id from the LangChainTrackingHandler
    for handler in callback_manager.handlers:
        if isinstance(handler, LangChainTrackingHandler):
            interaction = get_nearest_open_interaction()
            root_run_id = interaction._events_storage.get_root_id(  # pylint: disable=protected-access  # noqa: E501
                parent_run_id
            )
            additional_kwargs["root_run_id"] = str(root_run_id)
            break

    return additional_kwargs


def _process_prompt_template(
    inputs: dict[str, Any] | Any, prompt: PromptTemplate
) -> str:
    if isinstance(inputs, dict):
        return prompt.format(**{key: inputs.get(key) for key in prompt.input_variables})
    return prompt.format(**{prompt.input_variables[0]: inputs})


def _process_chat_prompt_template(
    inputs: dict[str, Any] | Any, prompt: ChatPromptTemplate
) -> tuple[str, list[HistoryEntry]]:
    messages = []
    for message in prompt.messages:
        if isinstance(inputs, dict):
            input_vars = {key: inputs.get(key) for key in message.input_variables}  # type: ignore  # noqa: E501  # pylint: disable=line-too-long
        else:
            input_vars = {message.input_variables[0]: inputs}  # type: ignore
        if isinstance(message, (HumanMessagePromptTemplate, AIMessagePromptTemplate)):
            messages.append((message.format(**input_vars).content))
    last_prompt = messages[-1]
    message_history = messages[:-1]

    if len(message_history) % 2 != 0:
        logger.warning("Odd number of chat history elements, ignoring last element")
        message_history = message_history[:-1]

    # Convert the history to [(user, assistant), ...] format
    history = [
        HistoryEntry(user=message_history[i], assistant=message_history[i + 1])
        for i in range(0, len(message_history), 2)
        if i < len(message_history) - 1
    ]

    return last_prompt, history


def _get_input_and_history(chain: Chain, inputs: dict[str, Any] | Any) -> ModelInput:
    chains = getattr(chain, "chains", None)
    if chains is not None:
        # If the chain is a SequentialChain, we need to get the
        # prompt from the first chain
        prompt = getattr(chains[0], "prompt", None)
    else:
        # If the chain is not a SequentialChain, we need to get
        # the prompt from the chain
        prompt = getattr(chain, "prompt", None)
        if prompt is None:
            if not isinstance(inputs, dict):
                return ModelInput(prompt=inputs)
            return ModelInput(prompt=inputs["input"])

    if isinstance(prompt, PromptTemplate):
        return ModelInput(prompt=_process_prompt_template(inputs, prompt))

    if isinstance(prompt, ChatPromptTemplate):
        prompt, history = _process_chat_prompt_template(inputs, prompt)
        return ModelInput(prompt=prompt, history=history)

    raise ValueError(f"Unknown prompt type: {prompt}")


def _get_output_chain(chain: Chain, result: dict[str, Any]) -> str:
    if len(chain.output_keys) == 1:
        return str(result[chain.output_keys[0]])
    output = {}
    for key in chain.output_keys:
        output[key] = result[key]
    return "\n".join([f"{key}: {value}" for key, value in output.items()])


def wrap_langchain(
    observer: Observer,
    function_name: str,
    f: Callable[[Any], Any],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    if function_name.startswith(
        ("llms.base.BaseLLM", "chat_models.base.BaseChatModel")
    ):
        additional_kwargs = _get_tracking_info_for_provider_call(**kwargs)
        return f(*args, **kwargs, **additional_kwargs)
    if function_name.startswith("chains.base.Chain"):
        try:
            get_nearest_open_interaction()
            return f(*args, **kwargs)
        except NotInInteractionContext:
            with new_interaction() as interaction:
                inputs = kwargs.get("inputs")
                handler = [
                    handler
                    for handler in kwargs.get("callbacks", [])
                    if isinstance(handler, LangChainTrackingHandler)
                ][0]
                interaction._set_user(  # pylint: disable=protected-access
                    handler.nebuly_user  # type: ignore
                )
                interaction._set_user_group_profile(  # pylint: disable=protected-access
                    handler.nebuly_user_group
                )
                interaction._set_observer(observer)  # pylint: disable=protected-access
                model_input = _get_input_and_history(args[0], inputs)
                interaction.set_input(model_input.prompt)
                interaction.set_history(model_input.history)
                original_res = f(*args, **kwargs)
                interaction.set_output(_get_output_chain(args[0], original_res))
                return original_res
    if function_name.startswith("indexes.vectorstore.VectorStoreIndexWrapper"):
        try:
            get_nearest_open_interaction()
            return f(*args, **kwargs)
        except NotInInteractionContext:
            with new_interaction() as interaction:
                inputs = get_argument(args, kwargs, "question", 1)
                interaction._set_user(  # pylint: disable=protected-access
                    kwargs.pop("user_id")
                )
                interaction._set_user_group_profile(  # pylint: disable=protected-access
                    kwargs.pop("user_group_profile", None)
                )
                interaction._set_observer(observer)  # pylint: disable=protected-access
                model_input = _get_input_and_history(args[0], inputs)
                interaction.set_input(model_input.prompt)
                interaction.set_history(model_input.history)
                original_res = f(*args, **kwargs)
                interaction.set_output(original_res)
                return original_res

    raise ValueError(f"Unknown function name: {function_name}")


async def wrap_langchain_async(
    observer: Observer,
    function_name: str,
    f: Callable[[Any], Any],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    if function_name.startswith(
        ("llms.base.BaseLLM", "chat_models.base.BaseChatModel")
    ):
        additional_kwargs = _get_tracking_info_for_provider_call(**kwargs)
        if isasyncgenfunction(f):
            return f(*args, **kwargs, **additional_kwargs)
        return await f(*args, **kwargs, **additional_kwargs)
    if function_name.startswith("chains.base.Chain"):
        try:
            get_nearest_open_interaction()
            if isasyncgenfunction(f):
                return f(*args, **kwargs)
            return await f(*args, **kwargs)
        except NotInInteractionContext:
            with new_interaction() as interaction:
                inputs = kwargs.get("inputs")
                if isinstance(inputs, dict):
                    user = inputs.pop("user_id", None)
                    user_group = inputs.pop("user_group_profile", None)
                    interaction._set_user(user)  # pylint: disable=protected-access
                    interaction._set_user_group_profile(  # pylint: disable=protected-access  # noqa: E501
                        user_group
                    )
                interaction._set_observer(observer)  # pylint: disable=protected-access
                model_input = _get_input_and_history(args[0], inputs)
                interaction.set_input(model_input.prompt)
                interaction.set_history(model_input.history)
                if isasyncgenfunction(f):
                    original_res = f(*args, **kwargs)
                else:
                    original_res = await f(*args, **kwargs)
                interaction.set_output(original_res)
                return original_res



from __future__ import annotations

import json
import logging
import os
import urllib.request
from time import sleep
from typing import Any
from unittest.mock import Mock

from nebuly.entities import InteractionWatch
from nebuly.exceptions import InvalidNebulyKeyError

logger = logging.getLogger(__name__)


class CustomJSONEncoder(json.JSONEncoder):
    ACCEPTED_KEYS = ["_model_id", "_model", "_context", "_examples", "_message_history"]

    @staticmethod
    def _check_key(key: str) -> bool:
        """
        Reject all private keys except for ones that contain info about the model
        """
        return not key.startswith("_") or key in CustomJSONEncoder.ACCEPTED_KEYS

    @staticmethod
    def _check_value(value: Any) -> bool:
        correct_types: tuple[type[Any], ...] = (
            str,
            int,
            float,
            bool,
            list,
            tuple,
            dict,
            type(None),
        )
        has_correct_type = isinstance(value, correct_types)
        is_valid_class = False
        try:
            if hasattr(value, "__dict__"):
                json.dumps(value.__dict__)
                is_valid_class = True
        except TypeError:
            is_valid_class = not isinstance(value, Mock) and any(
                (hasattr(value, key) for key in CustomJSONEncoder.ACCEPTED_KEYS)
            )
        return has_correct_type or is_valid_class

    def default(self, o: Any) -> Any:
        if isinstance(o, bytes):
            return o.decode("utf-8")
        if hasattr(o, "to_dict"):
            return o.to_dict()
        try:
            return json.JSONEncoder.default(self, o)
        except Exception:  # pylint: disable=broad-except
            if hasattr(o, "__dict__"):
                dict_repr = o.__dict__
                return {
                    key: dict_repr[key]
                    for key in dict_repr
                    if self._check_key(key) and self._check_value(dict_repr[key])
                }
            return str(o)


def post_message(
    watched: InteractionWatch, api_key: str, anonymize: bool = True
) -> None:
    message = json.dumps(
        {"body": watched, "anonymize": anonymize}, cls=CustomJSONEncoder
    )
    url = os.environ.get(
        # TODO: make this url configurable
        "NEBULY_API_URL",
        "https://backend.nebuly.com/event-ingestion/api/v1/events",
    )
    post_json_data(url, message, api_key)


def post_json_data(url: str, json_data: str, api_key: str) -> Any:
    request = urllib.request.Request(
        url,
        data=json_data.encode("utf-8"),
        headers={
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        },
    )

    while True:
        try:
            with urllib.request.urlopen(request, timeout=3) as response:  # nosec
                response_body = response.read().decode("utf-8")
                logger.debug("response_body: %s", response_body)
                break
        except urllib.error.HTTPError as e:
            if e.code == 401:
                raise InvalidNebulyKeyError(
                    "Invalid Nebuly API key, please provide a valid one and try again."
                ) from e
            logger.exception(
                "HTTPError when publishing the interaction: %s, "
                "retrying in 3 seconds...",
                e,
            )
            sleep(3)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(
                "Error when publishing the interaction: %s, "
                "retrying in 3 seconds...",
                e,
            )
            sleep(3)

    return response_body



from __future__ import annotations


class NebulyException(Exception):
    """Base class for exceptions in this module."""


class NebulyHTTPError(NebulyException):
    """Exception raised for errors in the HTTP request."""


class AlreadyImportedError(NebulyException):
    """Exception raised when the ai package was imported before init."""


class NebulyAlreadyInitializedError(NebulyException):
    """Exception raised when init is called more than once."""


class APIKeyNotProvidedError(NebulyException):
    """Exception raised when the API key is not provided."""


class InvalidNebulyKeyError(NebulyException):
    """Exception raised when the API key is not valid."""


class MissingRequiredNebulyFieldError(NebulyException):
    """Exception raised when a required nebuly field is missing."""


class InteractionContextError(NebulyException):
    """Exception raised when the API key is not valid."""


class InteractionContextInitiationError(InteractionContextError):
    """Exception raised when the interaction context is not initialized properly."""


class AlreadyInInteractionContext(InteractionContextError):
    """Exception raised when one interaction context is initialized inside another"""


class NotInInteractionContext(InteractionContextError):
    """Exception raised when no interaction context can be found."""


class InteractionMustBeLocalVariable(InteractionContextError):
    """Exception raised when the interaction manager creation misses the 'as'"""
